use super::context::IdentContext;
use crate::fqpath::*;
use proc_macro2::{Ident, Span, TokenStream};
use quote::{quote, ToTokens};
use syn::DeriveInput;

pub fn json_loader(input: DeriveInput) -> TokenStream {
	let context = match IdentContext::from_derive(input) {
		Ok(context) => context,
		Err(stream) => return stream,
	};

	let error_type = define_error_type(&context);
	let asset_set = define_loading_type(&context);
	let index_type = define_index_type(&context);
	let plugin = define_plugin(&context);
	let loader = define_loader(&context);
	let load_handler = define_load_handler(&context);

	quote! {
		#error_type
		#asset_set
		#index_type
		#loader
		#load_handler
		#plugin
	}
}

pub fn default_asset_storage() -> TokenStream {
	Ident::new("AssetHandles", Span::call_site()).to_token_stream()
}

pub fn define_loading_type(
	IdentContext {
		asset_name,
		set_name,
		..
	}: &IdentContext,
) -> TokenStream {
	quote! {
		#[derive(#FQSerialize, #FQDeserialize)]
		#[serde(untagged)]
		enum #set_name {
			One(#asset_name),
			Many(#FQVec<#asset_name>)
		}
	}
}

pub fn define_index_type(
	IdentContext {
		vis,
		index_name,
		asset_name,
		..
	}: &IdentContext,
) -> TokenStream {
	quote! {
		#[derive(#FQDebug, #BevyTypePath, #BevyDeref, #BevyDerefMut, #BevyAsset)]
		#vis struct #index_name(pub #FQHashMap<String, #BevyHandle<#asset_name>>);
	}
}

pub fn define_error_type(IdentContext { error_name, .. }: &IdentContext) -> TokenStream {
	quote! {
		#[derive(#FQDebug)]
		pub enum #error_name {
			Io(#FQIOError),
			Json(#SerdeJsonError)
		}

		impl #FQDisplay for #error_name {
			fn fmt(&self, f: &mut #FQFormatter<'_>) -> #FQFormatResult {
				match self {
					#error_name::Io(err) => err.fmt(f),
					#error_name::Json(err) => err.fmt(f),
				}
			}
		}

		impl #FQError for #error_name {
			fn source(&self) -> #FQOption<&(dyn #FQError + 'static)> {
				match self {
					#error_name::Io(err) => Some(err),
					#error_name::Json(err) => Some(err),
				}
			}
		}

		impl #FQFrom<#FQIOError> for #error_name {
			fn from(value: #FQIOError) -> #error_name {
				#error_name::Io(value)
			}
		}

		impl #FQFrom<#SerdeJsonError> for #error_name {
			fn from(value: #SerdeJsonError) -> #error_name {
				#error_name::Json(value)
			}
		}
	}
}

pub fn define_loader(
	IdentContext {
		vis,
		loader_name,
		index_name,
		set_name,
		error_name,
		id_field,
		extension,
		..
	}: &IdentContext,
) -> TokenStream {
	quote! {
		#[derive(#FQDefault)]
		#vis struct #loader_name;
		impl #BevyAssetLoader for #loader_name {
			type Asset = #index_name;
			type Settings = ();
			type Error = #error_name;

			async fn load<'a>(
				&'a self,
				mut reader: &'a mut #BevyAssetReader<'_>,
				_settings: &'a Self::Settings,
				load_context: &'a mut #BevyLoadContext<'_>,
			) -> #FQResult<Self::Asset, Self::Error> {
				let mut bytes = #FQVec::new();
				#BevyAsyncRead::read_to_end(&mut reader, &mut bytes).await?;
				let data: #set_name = ::serde_json::from_slice(bytes.as_slice())?;

				let mut asset_map = #FQHashMap::new();
				match data {
					#set_name::One(single_asset) => {
						let asset_id = format!("{}", &single_asset.#id_field);
						let handle = load_context.add_labeled_asset(
							asset_id.clone(),
							single_asset,
						);
						asset_map.insert(asset_id, handle);
					}
					#set_name::Many(asset_list) => {
						for single_asset in asset_list.into_iter() {
							let asset_id = format!("{}", &single_asset.#id_field);
							let handle = load_context.add_labeled_asset(
								asset_id.clone(),
								single_asset,
							);
							asset_map.insert(asset_id, handle);
						}
					}
				}
				Ok(#index_name(asset_map))
			}

			fn extensions(&self) -> &[&str] {
				&[#extension]
			}
		}
	}
}

pub fn define_plugin(
	IdentContext {
		vis,
		asset_name,
		index_name,
		loader_name,
		plugin_name,
		handler_name,
		..
	}: &IdentContext,
) -> TokenStream {
	quote! {
		#vis struct #plugin_name;
		impl #BevyPlugin for #plugin_name {
			fn build(&self, app: &mut #BevyApp) {
				#BevyAddAsset::init_asset::<#asset_name>(app);
				#BevyAddAsset::init_asset::<#index_name>(app);
				#BevyAddAsset::init_asset_loader::<#loader_name>(app);
				app.add_systems(#BevyUpdate, #handler_name);
			}
		}
	}
}

pub fn define_load_handler(
	IdentContext {
		vis,
		index_name,
		handler_name,
		storage_type,
		storage_asset_name,
		storage_index_name,
		..
	}: &IdentContext,
) -> TokenStream {
	quote! {
		#vis fn #handler_name(
			mut events: #BevyEventReader<#BevyAssetEvent<#index_name>>,
			mut asset_data: #BevyRes<#BevyAssets<#index_name>>,
			mut asset_storage: #BevyResMut<#storage_type>,
		) {
			let mut removed_containers = #FQVec::with_capacity(4);
			let mut removed_assets = #FQVec::with_capacity(4);

			for event in events.read() {
				match event {
					#BevyAssetEvent::LoadedWithDependencies { id } | #BevyAssetEvent::Added { id } | #BevyAssetEvent::Modified { id } => {
						let handle = #BevyHandle::Weak(*id);
						if let Some(asset_container) = asset_data.get(&handle) {
							for (id, handle) in asset_container.iter() {
								asset_storage.#storage_asset_name.insert(id.clone(), handle.clone());
							}
						}
					}
					#BevyAssetEvent::Removed { id } => {
						asset_storage.#storage_index_name.iter().for_each(|(key, stored)| {
							if stored.id() == *id {
								if let Some(asset_container) = asset_data.get(stored) {
									for (key, _) in asset_container.iter() {
										removed_assets.push(key.clone());
									}
									removed_containers.push(key.clone());
								}
							}
						});
					}
					#BevyAssetEvent::Unused { .. } => {}
				}
			}

			for id in removed_containers {
				asset_storage.#storage_index_name.remove(&id);
			}

			for id in removed_assets {
				asset_storage.#storage_asset_name.remove(&id);
			}
		}
	}
}
