use super::components::default_asset_storage;
use crate::utils::{find_asset_id, find_attribute, ident_suffix, simple_attr_kv, snake_case_ident};
use proc_macro2::{Ident, TokenStream};
use quote::quote_spanned;
use syn::{spanned::Spanned, Data, DeriveInput, Visibility};

pub struct IdentContext {
	pub vis: Visibility,
	// -- Type names
	pub asset_name: Ident,
	pub index_name: Ident,
	pub set_name: Ident,
	pub plugin_name: Ident,
	pub loader_name: Ident,
	pub handler_name: Ident,
	pub error_name: Ident,
	// -- Property names
	pub id_field: Ident,
	pub storage_type: TokenStream,
	pub storage_asset_name: Ident,
	pub storage_index_name: Ident,
	// -- Data Values
	pub uuid: TokenStream,
	pub extension: TokenStream,
}

macro_rules! span_err {
    ($($tok:tt)*) => {Err(quote_spanned!($($tok)*))};
}

macro_rules! parse_macro_input_err {
	($tokenstream:ident as $ty:ty) => {
		match ::syn::parse::<$ty>($tokenstream) {
			Ok(data) => data,
			Err(err) => {
				return Err(proc_macro2::TokenStream::from(err.to_compile_error()));
			}
		}
	};
}

impl IdentContext {
	pub fn from_derive(input: DeriveInput) -> Result<Self, TokenStream> {
		let data = match input.data {
			Data::Struct(input_struct) => input_struct,
			Data::Enum(_) => {
				return span_err!(
					input.ident.span() =>
					compile_error!("Can't create JSON Loader for Enum type");
				);
			}
			Data::Union(_) => {
				return span_err!(
					input.ident.span() =>
					compile_error!("Can't create JSON Loader for Union type");
				);
			}
		};

		let loader_attr = match find_attribute(&input.attrs, "loader") {
			Some(attribute) => attribute,
			None => {
				return span_err!(
					input.ident.span() =>
					compile_error!("JSON assets requires a 'loader' attribute that specifies a file extension");
				);
			}
		};

		let attrs = simple_attr_kv(&loader_attr);

		if !attrs.contains_key("extension") {
			return span_err!(
				loader_attr.span() =>
				compile_error!("JSON assets must specify a file extension, in the form '#[loader(extension = \"json\")]'");
			);
		}
		if !attrs.contains_key("uuid") {
			return span_err!(
				loader_attr.span() =>
				compile_error!("JSON assets must specify a uuid for the generated index asset, in the form '#[loader(uuid = \"asset-uuid\")]'");
			);
		}

		let storage_type = match attrs
			.get("storage")
			.map(|storage| storage.parse::<TokenStream>())
		{
			Some(Ok(storage)) => storage,
			Some(Err(_err)) => {
				return span_err!(
					loader_attr.span() =>
					compile_error!("Invalid storage type, must provide a valid path");
				);
			}
			None => default_asset_storage(),
		};

		let DeriveInput { vis, ident, .. } = input;

		Ok(IdentContext {
			vis,

			// Type Names
			index_name: ident_suffix(&ident, "Index"),
			loader_name: ident_suffix(&ident, "Loader"),
			set_name: ident_suffix(&ident, "Set"),
			plugin_name: ident_suffix(&ident, "Plugin"),
			error_name: ident_suffix(&ident, "Error"),
			handler_name: ident_suffix(&snake_case_ident(&ident), "_load_handler"),
			asset_name: ident.clone(),

			// Property Names
			id_field: match find_asset_id(&data) {
				Ok(field) => match field.ident {
					Some(ident) => ident,
					None => {
						return span_err!(
							field.span() =>
							compile_error!("asset_id field must have a name");
						);
					}
				},
				Err(Some(stream)) => {
					return Err(stream);
				}
				Err(None) => {
					return span_err!(
						ident.span() =>
						compile_error!("Asset must contain either a property named 'id', or a named property labelled with the 'asset_id' attribute");
					);
				}
			},
			storage_type,
			storage_index_name: match attrs.get("index_name") {
				Some(index_property) => match index_property.parse::<TokenStream>() {
					Ok(ident) => {
						let ident = proc_macro::TokenStream::from(ident);
						parse_macro_input_err!(ident as Ident)
					}
					Err(_) => {
						return Err(quote_spanned!(
							loader_attr.span() =>
							compile_error!("Failed to parse 'index_name' attribute");
						));
					}
				},
				None => ident_suffix(&snake_case_ident(&ident), "_index"),
			},
			storage_asset_name: match attrs.get("asset_name") {
				Some(asset_property) => match asset_property.parse::<TokenStream>() {
					Ok(ident) => {
						let ident = proc_macro::TokenStream::from(ident);
						parse_macro_input_err!(ident as Ident)
					}
					Err(_) => {
						return span_err!(
							loader_attr.span() =>
							compile_error!("Failed to parse 'asset_name' attribute");
						);
					}
				},
				None => snake_case_ident(&ident),
			},

			// -- Data Values
			uuid: match attrs["uuid"].parse::<TokenStream>() {
				Ok(value) => value,
				Err(_) => {
					return span_err!(
						loader_attr.span() =>
						compile_error!("Invalid uuid definition");
					);
				}
			},
			extension: match attrs["extension"].parse::<TokenStream>() {
				Ok(value) => value,
				Err(_) => {
					return span_err!(
						loader_attr.span() =>
						compile_error!("Invalid extension definition");
					);
				}
			},
		})
	}
}
