use crate::fqpath::{
	BevyComponent, BevyEntity, BevyQuery, BevySystemParam, BevyWith, FQClone, FQCopy, FQDebug,
	FQDefault,
};
use crate::utils::{ident_prefix, only_named_struct, only_struct};
use proc_macro2::TokenStream;
use quote::{quote, quote_spanned};
use syn::spanned::Spanned;
use syn::{DeriveInput, Field};

macro_rules! or_return {
	($expr: expr) => {
		match $expr {
			Ok(value) => value,
			Err(err) => return err,
		}
	};
}

pub fn tag_finder(input: DeriveInput) -> TokenStream {
	let struct_data = or_return!(only_struct(&input));
	let _ = or_return!(only_named_struct(struct_data));

	let struct_fields: TokenStream = struct_data.fields.iter().map(map_named_field).collect();
	let struct_methods: TokenStream = struct_data.fields.iter().map(map_named_method).collect();
	let tag_types: TokenStream = struct_data.fields.iter().map(map_tag_types).collect();

	let DeriveInput { vis, ident, .. } = input;

	quote! {
		#tag_types

		#[derive(#BevySystemParam)]
		#vis struct #ident <'w, 's> {
			#struct_fields
		}

		impl <'w, 's> #ident<'w, 's> {
			#struct_methods
		}
	}
}

fn map_tag_types(field: &Field) -> TokenStream {
	let Field { ty, attrs, .. } = field;

	let is_sparse = attrs.iter().any(|attr| attr.path().is_ident("sparse"));

	if is_sparse {
		quote! {
			#[derive(#FQCopy, #FQClone, #FQDefault, #FQDebug, #BevyComponent)]
			#[component(storage = "SparseSet")]
			pub struct #ty;
		}
	} else {
		quote! {
			#[derive(#FQCopy, #FQClone, #FQDefault, #FQDebug, #BevyComponent)]
			pub struct #ty;
		}
	}
}

fn map_named_field(field: &Field) -> TokenStream {
	let Field { ident, ty, .. } = field;
	quote! {
		#ident: #BevyQuery<'w, 's, (), #BevyWith<#ty>>,
	}
}

fn map_named_method(field: &Field) -> TokenStream {
	let Field { ident, ty, .. } = field;
	match ident {
		None => {
			quote_spanned! { ty.span() => 	compile_error!("Unable to map field with type {} that does not have a name"); }
		}
		Some(ident) => {
			let name = ident_prefix(ident, "is_");
			quote! {
				pub fn #name(&self, entity: #BevyEntity) -> bool {
					self.#ident.contains(entity)
				}
			}
		}
	}
}
